use std::collections::HashMap;
use std::env;
use std::ffi::OsString;
use std::mem::forget;
use std::path::{Path, PathBuf};
use std::sync::atomic::{AtomicUsize, Ordering};
use std::sync::Arc;

use dunce::canonicalize;
use json::stringify;
use parking_lot::Mutex;

use pi_atom::Atom;
use pi_file::fs_monitor::{FSChangeEvent, FSListener, FSMonitor, FSMonitorOptions};
use pi_serv_lib::js_gray::{get_vid_contexts, GRAY_MGR};
use pi_serv_lib::js_hotfix_files::{hotfix_js_file, HOTFIX_FILES_MANAGER};
use vm_builtin::task_pool::VmTaskPool;
use vm_builtin::{ContextHandle, Pid, ProcessMsg};
use vm_core::vm::send_to_process;

use crate::js_net::HTTP_STATIC_CACHES;
use crate::MAIN_ASYNC_RUNTIME;
use crate::VID_CONTEXTS;

const INIT_FILE: &str = "pi_pt/init.js";

/// 热更配置文件
fn hotfix_cfg(change_path: PathBuf, prefix: PathBuf, gray: usize) {
    let path = canonicalize(&change_path).unwrap();
    let path = path
        .strip_prefix(&prefix)
        .unwrap()
        .to_str()
        .unwrap()
        .to_string();
    let path = path.replace("\\", "/");

    // init.js 的改变不应该热更，因为init.js中会启动网络服务，导致冲突
    if path == INIT_FILE {
        return;
    }

    let vsn = HOTFIX_FILES_MANAGER.add(gray, path.clone());
    if let Some(vids) = GRAY_MGR.read().gray_vids(gray) {
        if let Some(vid) = vids.first() {
            let change_path = change_path.clone();
            let path = path.clone();
            let pid = Pid(vid.clone(), 0);
            let mut msgs = Vec::with_capacity(3);
            msgs.push(ProcessMsg::String("hotfix_cfg".to_string()));
            msgs.push(ProcessMsg::String(path));
            msgs.push(ProcessMsg::Number(vsn as f64));
            // 管理PID发送消息
            send_to_process::<VmTaskPool, VmTaskPool>(None, pid, ProcessMsg::Array(msgs));
        }
    }
}

/// 热更接口文件
fn hotfix_topic(change_path: PathBuf, prefix: PathBuf, gray: usize) {
    let path = canonicalize(&change_path).unwrap();
    let path = path
        .strip_prefix(&prefix)
        .unwrap()
        .to_str()
        .unwrap()
        .to_string();
    let path = path.replace("\\", "/");

    // init.js 的改变不应该热更，因为init.js中会启动网络服务，导致冲突
    if path == INIT_FILE {
        return;
    }

    let vsn = HOTFIX_FILES_MANAGER.add(gray, path.clone());
    if let Some(vids) = GRAY_MGR.read().gray_vids(gray) {
        if let Some(vid) = vids.first() {
            let change_path = change_path.clone();
            let path = path.clone();
            let pid = Pid(vid.clone(), 0);
            let mut msgs = Vec::with_capacity(3);
            msgs.push(ProcessMsg::String("hotfix_topic".to_string()));
            msgs.push(ProcessMsg::String(path));
            msgs.push(ProcessMsg::Number(vsn as f64));
            // 管理PID发送消息
            send_to_process::<VmTaskPool, VmTaskPool>(None, pid, ProcessMsg::Array(msgs));
        }
    }
}

/// 热更代码文件
fn module_changed(change_path: PathBuf, prefix: PathBuf, gray: usize) {
    let path = canonicalize(&change_path).unwrap();
    let path = path
        .strip_prefix(&prefix)
        .unwrap()
        .to_str()
        .unwrap()
        .to_string();
    let path = path.replace("\\", "/");

    // init.js 的改变不应该热更，因为init.js中会启动网络服务，导致冲突
    if path == INIT_FILE {
        return;
    }
    let vsn = HOTFIX_FILES_MANAGER.add(gray, path.clone());

    if let Some(vms) = GRAY_MGR.read().get_vms(gray) {
        for vm in vms {
            if let Some(ctxs) = get_vid_contexts(vm.get_vid()) {
                // 虚拟机的每个context重新require该模块
                for context in ctxs {
                    hotfix_js_file(vm.clone(), context, path.clone(), vsn);
                }
            }
        }
    }
}

// 删除模块
fn module_removed(change_path: PathBuf, gray: usize) {
    let path = change_path.to_str().unwrap();
    let path = path.replace("\\", "/");

    if let Some(vms) = GRAY_MGR.read().get_vms(gray) {
        for vm in vms {
            if let Some(ctxs) = get_vid_contexts(vm.get_vid()) {
                for context in ctxs {
                    let vm_copy = vm.clone();
                    let vm = vm.clone();
                    let path = path.clone();
                    let context = context.clone();
                    vm_copy.spawn_task(async move {
                        if let Ok(Some(func)) = vm
                            .get_property(context.clone(), "self.Module.delete_module")
                            .await
                        {
                            let path_js = vm
                                .to_js_value(context.clone(), stringify(path.clone()))
                                .await
                                .unwrap()
                                .unwrap();
                            if let Err(_e) = vm.call(context.clone(), &func, vec![path_js]).await {
                                warn!("hotfix call delete_module error");
                            } else {
                                debug!("module_removed success path: {:?}", path);
                            }
                        } else {
                            warn!("hotfix get vm context failed");
                        }
                    });
                }
            }
        }
    }
}

/// 后端资源热更
pub fn hotfix_listen_backend(gray: usize, project_path: String) {
    let current_dir = env::current_dir().unwrap();
    let current_dir_parent = current_dir.parent().unwrap().to_str().unwrap();
    let path = Path::new(&project_path)
        .iter()
        .filter_map(|x| if x == "." || x == ".." { None } else { Some(x) })
        .map(|x| x.to_str().unwrap())
        .collect::<Vec<&str>>();
    let root: PathBuf = [vec![current_dir_parent], path].concat().iter().collect();
    let project_root = root.to_str().unwrap().replace("\\", "/");
    let listener = FSListener(Arc::new(move |event: FSChangeEvent| {
        let prefix = canonicalize(project_root.clone()).unwrap();

        match event {
            FSChangeEvent::Create(path) => {
                let path_clone = path.clone();
                path.extension().and_then(|p| {
                    p.to_str().and_then(|s| {
                        if s == "js" {
                            debug!("hotfix create path {:?}", path);
                            let full_exten = get_full_extension(&path);
                            if Some("topic.js".to_string()) == full_exten {
                                // 热更接口
                                Some(hotfix_topic(path_clone, prefix, gray))
                            } else {
                                // 热更worker虚拟机
                                Some(module_changed(path_clone.clone(), prefix.clone(), gray))
                            }
                        } else if s == "bon" {
                            debug!("hotfix cfg create path {:?}", path);
                            Some(hotfix_cfg(path_clone, prefix, gray))
                        } else {
                            None
                        }
                    })
                });
            }
            FSChangeEvent::Write(path) => {
                let path_clone = path.clone();
                path.extension().and_then(|p| {
                    p.to_str().and_then(|s| {
                        if s == "js" {
                            debug!("hotfix modified path {:?}", path);
                            let full_exten = get_full_extension(&path);
                            if Some("topic.js".to_string()) == full_exten {
                                // 热更接口
                                Some(hotfix_topic(path_clone, prefix, gray))
                            } else {
                                Some(module_changed(path_clone.clone(), prefix.clone(), gray))
                            }
                        } else if s == "bon" {
                            debug!("hotfix cfg create path {:?}", path);
                            Some(hotfix_cfg(path_clone, prefix, gray))
                        } else {
                            None
                        }
                    })
                });
            }
            FSChangeEvent::Remove(path) => {
                let path_clone = path.clone();
                path.extension().and_then(|p| {
                    p.to_str().and_then(|s| {
                        if s == "js" {
                            debug!("hotfix remove path {:?}", path);
                            Some(module_removed(path_clone.clone(), gray))
                        } else {
                            None
                        }
                    })
                });
            }
            FSChangeEvent::Rename(old, new) => {
                let path_clone = new.clone();
                new.extension().and_then(|p| {
                    p.to_str().and_then(|s| {
                        if s == "js" {
                            debug!("hotfix rename path old = {:?}, new = {:?}", old, new);
                            Some(module_changed(path_clone.clone(), prefix.clone(), gray));
                            Some(module_removed(path_clone.clone(), gray))
                        } else {
                            None
                        }
                    })
                });
            }
        };
    }));
    let mut monitor = FSMonitor::new(
        FSMonitorOptions::Dir(Atom::from(project_path), true, 1000),
        listener,
    );
    monitor.run().expect("watch dir failed");
    forget(monitor);
}

/// 前端资源热更， 清除http缓存
pub fn hotfix_listen_frontend() {
    for cache in HTTP_STATIC_CACHES.read().iter() {
        cache.remove_all_cache();
    }
}

fn get_full_extension(path: &Path) -> Option<String> {
    if let Some(file_name) = path.file_name() {
        let file_name = file_name.to_string_lossy();
        let parts: Vec<&str> = file_name.split('.').collect();
        if parts.len() > 1 {
            return Some(parts[1..].join("."));
        }
    }
    None
}
