use parking_lot::RwLock;
use std::io::{Error, ErrorKind, Result};
use std::path::PathBuf;
use std::sync::Arc;

use futures::future::{LocalBoxFuture, FutureExt};

use log::warn;
use pi_handler::{Args, Handler, SGenType};
use pi_hash::XHashMap;
use tcp::Socket;

use dashmap::DashMap;
use http::{
    gateway::GatewayContext,
    middleware::{Middleware, MiddlewareResult},
    request::HttpRequest,
    response::HttpResponse,
    utils::{trim_path, HttpRecvResult},
};

use safety_sign::kms::{self, check_input, restore_secret};

/*
* https文件上传任务优先级
*/
const HTTPS_ASYNC_FILE_UPLOAD_PRIORITY: usize = 100;

/// 分发秘钥
const SET_RANDOM_METHOD: &str = "set_random";
/// 生成私钥
const CREATE_KEY_METHOD: &str = "create_key";
/// 设置秘钥
const SET_KEYS_METHOD: &str = "set_keys";

/*
* 安全密钥处理器
*/
pub struct SafetySign {
    random_map: Arc<DashMap<String, Vec<(String, String)>>>,
    input_key: Arc<DashMap<String, Vec<String>>>,
}

unsafe impl Send for SafetySign {}

unsafe impl Sync for SafetySign {}

impl<S: Socket> Middleware<S, GatewayContext> for SafetySign {
    fn request<'a>(
        &'a self,
        context: &'a mut GatewayContext,
        req: HttpRequest<S>,
    ) -> LocalBoxFuture<'a, MiddlewareResult<S>> {
        let resp = HttpResponse::new(2);
        let resp_handler = resp.get_response_handler().unwrap();
        // let map = context.as_mut_parts();

        let args = context.as_params().clone();
        let map = &*args.borrow();

        let mut key_cp = None;
        if let Some(SGenType::Str(key)) = map.get("key") {
            key_cp = Some(key.clone());
        };

        let mut project_cp = None;
        if let Some(SGenType::Str(key)) = map.get("project") {
            project_cp = Some(key.clone());
        };

        let mut api_cp = None;
        if let Some(SGenType::Str(key)) = map.get("api") {
            api_cp = Some(key.clone());
        };
        let mut random_cp = None;
        if let Some(SGenType::Str(key)) = map.get("random") {
            random_cp = Some(key.clone());
        };
        let mut index_cp = None;
        if let Some(SGenType::Str(key)) = map.get("index") {
            index_cp = Some(key.clone());
        };

        let future = async move {
            /// 随机数
            let mut random = "".to_string();
            /// 项目
            let mut project = "".to_string();

            if let Some(var) = project_cp {
                project = var.clone();
            } else {
                return MiddlewareResult::Throw(Error::new(
                    ErrorKind::NotFound,
                    "arg project null",
                ));
            }
            if let Some(api) = api_cp {
                match api.as_str() {
                    SET_RANDOM_METHOD => {
                        let index;
                        if let Some(var) = random_cp {
                            random = var.clone();
                        } else {
                            return MiddlewareResult::Throw(Error::new(
                                ErrorKind::NotFound,
                                "arg random null",
                            ));
                        }
                        if let Some(var) = index_cp {
                            index = var.parse::<usize>().unwrap();
                            if (index > 4 || index < 0) {
                                return MiddlewareResult::Throw(Error::new(
                                    ErrorKind::NotFound,
                                    "arg index 0-4",
                                ));
                            }
                        } else {
                            return MiddlewareResult::Throw(Error::new(
                                ErrorKind::NotFound,
                                "arg index null",
                            ));
                        }
                        let _ = set_random_handle(
                            self.random_map.clone(),
                            &resp,
                            project.clone(),
                            random,
                            index,
                        )
                        .await;
                    }
                    SET_KEYS_METHOD => {
                        if let Some(key) = key_cp {
                            let map = &self.input_key;
                            // 检测秘钥是否已经导入
                            if check_input(project.clone()) {
                                if let Err(e) = resp_handler
                                    .write(Vec::from(format!("The imported ok").as_bytes()))
                                    .await
                                {
                                    warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
                                } else {
                                    if let Err(e) = resp_handler.finish().await {
                                        warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                                    }
                                }
                            } else {
                                if let Some(mut e) = map.get_mut(&project) {
                                    let keys = e.value_mut();
                                    if !keys.contains(&key) {
                                        keys.push(key.clone());
                                    }
                                    // map.insert(project.clone(), keys.to_vec());
                                    if keys.len() >= 3 {
                                        // 导入秘钥
                                        let _ =
                                            input_key_handle(&resp, project.clone(), keys.clone())
                                                .await;
                                    } else {
                                        if let Err(e) = resp_handler
                                            .write(Vec::from(format!("ok").as_bytes()))
                                            .await
                                        {
                                            warn!(
                                                "!!!> Http Body Mut Write Failed, reason: {:?}",
                                                e
                                            );
                                        } else {
                                            if let Err(e) = resp_handler.finish().await {
                                                warn!(
                                                "!!!> Http Body Mut Finish Failed, reason: {:?}",
                                                e
                                            );
                                            }
                                        }
                                    }
                                } else {
                                    let mut list = Vec::new();
                                    list.push(key.clone());
                                    map.insert(project.clone(), list);
                                    // 初始化数据成功
                                    if let Err(e) = resp_handler
                                        .write(Vec::from(format!("ok").as_bytes()))
                                        .await
                                    {
                                        warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
                                    } else {
                                        if let Err(e) = resp_handler.finish().await {
                                            warn!(
                                                "!!!> Http Body Mut Finish Failed, reason: {:?}",
                                                e
                                            );
                                        }
                                    }
                                }
                            }
                        } else {
                            return MiddlewareResult::Throw(Error::new(
                                ErrorKind::NotFound,
                                "arg key null",
                            ));
                        }
                    }
                    _ => {
                        return MiddlewareResult::Throw(Error::new(
                            ErrorKind::NotFound,
                            "arg api error",
                        ));
                    }
                }
            }
            //完成请求处理
            MiddlewareResult::Finish((req, resp))
            // MiddlewareResult::Break(resp)
        };
        future.boxed_local()
    }

    fn response<'a>(
        &'a self,
        context: &'a mut GatewayContext,
        req: HttpRequest<S>,
        resp: HttpResponse,
    ) -> LocalBoxFuture<'a, MiddlewareResult<S>> {
        let mut body_bufs: Vec<Vec<u8>> = Vec::new();
        let mut response = resp;
        let future = async move {
            if let Some(body) = response.as_mut_body() {
                //当前响应有响应体，则持续获取响应体的内容
                loop {
                    match body.body().await {
                        HttpRecvResult::Err(e) => {
                            //获取Http响应体错误
                            return MiddlewareResult::Throw(e);
                        }
                        HttpRecvResult::Ok(bodys) => {
                            //获取到的是Http响应体块的后继
                            for (_index, bin) in bodys {
                                body_bufs.push(bin);
                            }
                        }
                        HttpRecvResult::Fin(bodys) => {
                            //获取到的是Http响应体块的尾部，处理后退出循环
                            body.init(); //未初始化，则初始化响应体
                            for buf in body_bufs {
                                body.push(buf.as_slice());
                            }
                            for (_index, bin) in bodys {
                                body.push(bin.as_slice());
                            }
                            break;
                        }
                    }
                }
            }
            //继续响应处理
            MiddlewareResult::ContinueResponse((req, response))
        };
        future.boxed_local()
    }
}

impl SafetySign {
    //构建指定根目录的文件上传处理器
    pub fn new() -> Self {
        SafetySign {
            random_map: Arc::new(DashMap::default()),
            input_key: Arc::new(DashMap::default()),
        }
    }
}

/// 设置随机数
async fn set_random_handle(
    data: Arc<DashMap<String, Vec<(String, String)>>>,
    resp: &HttpResponse,
    project: String,
    random: String,
    index: usize,
) -> Result<()> {
    if let Some(resp_handler) = resp.get_response_handler() {
        // 判断项目是否存在
        if data.get(&project).is_none() {
            let mut list = vec![("".to_string(), "".to_string()); 5];
            list[index] = (random, "".to_string());
            data.insert(project, list);
            // 初始化数据成功
            if let Err(e) = resp_handler
                .write(Vec::from(
                    format!("Number of imported random: 1").as_bytes(),
                ))
                .await
            {
                warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
            } else {
                if let Err(e) = resp_handler.finish().await {
                    warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                }
            }
        } else {
            let mut map = data;
            if let Some(mut e) = map.get_mut(&project) {
                let list = e.value_mut();
                let (random2, key) = &list[index];
                if key.is_empty() {
                    list[index] = (random, "".to_string());
                    let _ = create_key_handle(list, &resp, project.clone()).await;
                } else if random == random2.clone() {
                    // 秘钥存在则发送
                    if let Err(e) = resp_handler
                        .write(Vec::from(format!("key:{:?}", key).as_bytes()))
                        .await
                    {
                        warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
                    } else {
                        if let Err(e) = resp_handler.finish().await {
                            warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                        }
                    }
                }
            };
        }
        return Ok(());
    }

    Err(Error::new(
        ErrorKind::Other,
        "error, reason: invalid response body",
    ))
}

/// 生成私钥
async fn create_key_handle(
    data: &mut Vec<(String, String)>,
    resp: &HttpResponse,
    project: String,
) -> Result<()> {
    if let Some(resp_handler) = resp.get_response_handler() {
        // 判断项目是否存在
        // if let Some(list) = data.read().get(&project) {
        let mut randoms = Vec::with_capacity(5);
        for i in 0..data.len() {
            let (r, _) = &data[i];
            if (r.as_str() != "") {
                randoms.push(r.clone());
            }
        }
        if randoms.len() == 5 {
            let entropy = format!(
                "{:?}{:?}{:?}{:?}{:?}",
                randoms[0], randoms[1], randoms[2], randoms[3], randoms[4]
            );
            let keys = kms::create_secret(entropy.as_bytes(), 5, 3);
            for x in 0..data.len() {
                if let Some((_, key)) = data.get_mut(x) {
                    *key = keys[x].clone();
                }
            }
            // let _ = resp_handler.write(Vec::from(format!("ok").as_bytes()));
            if let Err(e) = resp_handler
                .write(Vec::from(
                    format!("The key generation succeeded. Please refresh the page").as_bytes(),
                ))
                .await
            {
                warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
            } else {
                if let Err(e) = resp_handler.finish().await {
                    warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                }
            }
        } else {
            /// 导入随机数份数
            // let _ = resp_handler.write(Vec::from(
            //     format!("random count:{:?}", randoms.len()).as_bytes(),
            // ));
            if let Err(e) = resp_handler
                .write(Vec::from(
                    format!("Number of imported random:{:?}", randoms.len()).as_bytes(),
                ))
                .await
            {
                warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
            } else {
                if let Err(e) = resp_handler.finish().await {
                    warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                }
            }
        }
        return Ok(());
    }

    Err(Error::new(
        ErrorKind::Other,
        "error, reason: invalid response body",
    ))
}

/// 导入私钥
async fn input_key_handle(resp: &HttpResponse, project: String, keys: Vec<String>) -> Result<()> {
    if let Some(resp_handler) = resp.get_response_handler() {
        if restore_secret(project, &keys).is_ok() {
            // let _ = resp_handler.write(Vec::from(format!("ok").as_bytes()));
            if let Err(e) = resp_handler
                .write(Vec::from(
                    format!("Importing the key is complete").as_bytes(),
                ))
                .await
            {
                warn!("!!!> Http Body Mut Write Failed, reason: {:?}", e);
            } else {
                if let Err(e) = resp_handler.finish().await {
                    warn!("!!!> Http Body Mut Finish Failed, reason: {:?}", e);
                }
            }
        }
        return Ok(());
    }

    Err(Error::new(
        ErrorKind::Other,
        "error, reason: invalid response body",
    ))
}
