#![deny(unused_must_use)]
#![allow(unused_imports)]
#![allow(unused_variables)]

#[macro_use]
extern crate log;
#[macro_use]
extern crate lazy_static;
#[macro_use]
extern crate json;

#[cfg(feature = "default")]
#[macro_use]
extern crate pi_core;
#[cfg(feature = "jemalloc_profiling")]
#[macro_use]
extern crate jemalloc_pi_core;
#[cfg(feature = "profiling_heap")]
#[macro_use]
extern crate profiling_pi_core;

use std::collections::HashSet;
use std::path::{Path, PathBuf};
use std::sync::mpsc::{channel, Sender};
use std::sync::{
    atomic::{AtomicBool, AtomicUsize, Ordering},
    Arc,
};
use std::thread;
use std::time::{Duration, Instant};
use std::{env, fs::read_to_string};

use clap::{App, Arg, ArgMatches, SubCommand};
use json::{array, parse, stringify};
use license_client::license_handle;
use minstant;
use num_cpus::get_physical;
use parking_lot::{Condvar, Mutex, RwLock, WaitTimeoutResult};

use pi_async::rt::serial::AsyncRuntimeBuilder;
use pi_async::rt::serial_local_thread::LocalTaskRuntime;
use pi_async::rt::AsyncRuntimeExt;
use pi_async::rt::{
    multi_thread::{MultiTaskRuntime, MultiTaskRuntimeBuilder, StealableTaskPool},
    single_thread::{SingleTaskRunner, SingleTaskRuntime},
    spawn_worker_thread, startup_global_time_loop, AsyncRuntime,
};
use pi_hash::XHashMap;
use pi_serv_lib::js_gray::GrayVmArgs;
use vm_builtin::{
    task_pool::VmTaskPool, utils::set_used_rust_heap_size_getter, AsyncVmTaskPool, ContextHandle,
    Pid, ProcessMsg, VmEvent, VmEventHandler, VmEventValue, VmStartupSnapshot, VmTiming,
};
use vm_core::{debug, init_v8, vm, vm::send_to_process};

#[cfg(all(
    target_arch = "x86_64",
    target_os = "linux",
    feature = "jemalloc_profiling"
))]
use jemalloc_pi_core::{
    allocator::alloced_size,
    console::{set_console_shell_ctrlc_handler, ConsoleShell, ConsoleShellBuilder},
    create_snapshot_vm, finish_snapshot, init_snapshot, init_v8_env, init_work_vm,
    set_default_ctrlc_handler,
    terminal::VmTerminal,
};
#[cfg(feature = "default")]
use pi_core::{
    allocator::alloced_size,
    console::{set_console_shell_ctrlc_handler, ConsoleShell, ConsoleShellBuilder},
    create_snapshot_vm, finish_snapshot, init_snapshot, init_v8_env, init_work_vm,
    terminal::VmTerminal,
};
use pi_core_builtin::set_external_async_runtime;
use pi_core_lib::set_file_async_runtime;
use pi_serv_ext::register_ext_functions;
use pi_serv_lib::{
    js_db::{global_db_mgr, init_db},
    js_gray::{GRAY_MGR, VID_CONTEXTS},
    js_monitor,
};
use pi_serv_lib::{
    set_pi_serv_lib_db_runtime, set_pi_serv_lib_file_runtime, set_pi_serv_lib_main_async_runtime,
};
#[cfg(feature = "profiling_heap")]
use profiling_pi_core::{
    allocator::alloced_size,
    console::{set_console_shell_ctrlc_handler, ConsoleShell, ConsoleShellBuilder},
    create_snapshot_vm, finish_snapshot, init_snapshot, init_v8_env, init_work_vm,
    set_default_ctrlc_handler,
    terminal::VmTerminal,
};

mod hotfix;
mod http_safety_sign;
mod init;
mod js_net;
mod license_client;
mod mgr_pid;
mod pi_log;

use crate::js_net::create_listener_pid;
use crate::mgr_pid::mgr_pid_handle;
use hotfix::{hotfix_listen_backend, hotfix_listen_frontend};
use init::{init_js, init_main, read_init_source};
use js_net::{
    create_http_pid, reg_pi_serv_handle, set_init_state, start_network_services, PI_PT_INIT_STATE,
};
// use pi_db::db_collect::collect_db;
// use pi_db::log_file_db::LOG_FILE_SIZE;
use pi_time::run_millis;
use std::io::{Error, ErrorKind, Result};

#[cfg(feature = "default")]
init_global_counter_alloc! {}
#[cfg(all(
    target_arch = "x86_64",
    target_os = "linux",
    feature = "jemalloc_profiling"
))]
init_jemalloc_profiling_alloc! {}
#[cfg(feature = "profiling_heap")]
init_global_profiling_alloc! {}

lazy_static! {
    //主线程运行状态和线程无条件休眠超时时长
    static ref MAIN_RUN_STATUS: Arc<AtomicBool> = Arc::new(AtomicBool::new(true));
    static ref MAIN_UNCONDITION_SLEEP_TIMEOUT: u64 = 10;

    //主线程条件变量和线程条件休眠超时时长
    static ref MAIN_CONDVAR: Arc<(AtomicBool, Mutex<()>, Condvar)> = Arc::new((AtomicBool::new(false), Mutex::new(()), Condvar::new()));
    static ref MAIN_CONDITION_SLEEP_TIMEOUT: u64 = 10000;

    //初始化主线程异步运行时
    static ref MAIN_ASYNC_RUNNER: SingleTaskRunner<(), VmTaskPool> = SingleTaskRunner::new(VmTaskPool::default());
    static ref MAIN_ASYNC_RUNTIME: SingleTaskRuntime<(), VmTaskPool> = MAIN_ASYNC_RUNNER.startup().unwrap();

    //初始化文件异步运行时
    static ref FILES_ASYNC_RUNTIME: MultiTaskRuntime<()> = {
        let mut len = get_physical() + 1;
        if let Ok(var) = env::var("FILES_RUNTIME_COUNT") {
            len = var.parse().unwrap_or(10);
        }

        let pool = StealableTaskPool::with(len, 1000000, [1, 1], 3000);
        let builder = MultiTaskRuntimeBuilder::new(pool)
            .thread_prefix("PI-SERV-FILE")
            .thread_stack_size(2 * 1024 * 1024)
            .init_worker_size(len)
            .set_worker_limit(len, len)
            .set_timeout(10)
            .set_timer_interval(1);
        builder.build()
    };

    //初始化数据库异步运行时
    static ref DB_ASYNC_RUNTIME: MultiTaskRuntime<()> = {
        let mut len = get_physical() + 1;
        if let Ok(var) = env::var("DB_RUNTIME_COUNT") {
            len = var.parse().unwrap_or(10);
        }

        let pool = StealableTaskPool::with(len, 1000000, [1, 1], 3000);
        let builder = MultiTaskRuntimeBuilder::new(pool)
            .thread_prefix("PI-SERV-DB")
            .thread_stack_size(2 * 1024 * 1024)
            .init_worker_size(len)
            .set_worker_limit(len, len)
            .set_timeout(10)
            .set_timer_interval(1);
        builder.build()
    };

    //Mqtt端口代理映射表
    static ref MQTT_PORTS: Arc<Mutex<HashSet<(u16, String)>>> = Arc::new(Mutex::new(HashSet::new()));
    //Http端口代理映射表(端口,主机名,配置)
    static ref HTTP_PORTS: Arc<Mutex<HashSet<(u16, String, String)>>> = Arc::new(Mutex::new(HashSet::new()));

    //控制台
    static ref CONSOLE_SHELL: RwLock<Option<ConsoleShell<SingleTaskRuntime<(), VmTaskPool>,  VmTaskPool, VmTaskPool>>> = RwLock::new(None);

    //自动GC
    static ref AUTO_GC: RwLock<XHashMap<usize, AtomicUsize>> = RwLock::new(XHashMap::default());

    // project对应的灰度版本
    static ref PROJECT_GRAY: RwLock<XHashMap<String, usize>> = RwLock::new(XHashMap::default());

    // 动态加载版本
    static ref LOAD_PROJECT_SENDER: Arc<Mutex<Option<Sender<(Option<usize>, Option<usize>, Option<u16>, bool, usize, String, bool, String, usize, Option<Sender<bool>>)>>>> = Arc::new(Mutex::new(None));

}

/*
* 同步执行入口，退出时会中止主线程
*/
fn main() {
    //初始化分析堆内存的堆分配器
    #[cfg(feature = "profiling_heap")]
    init_profiling_alloctor();
    let _handle = startup_global_time_loop(10);
    // 开启防私服
    // license_handle();

    // 调用当前运行时间，如果不提前初始化，其他功能使用pi_lib中的time时候可能会异常
    run_millis();

    set_used_rust_heap_size_getter(alloced_size);

    //初始化日志服务器
    let _ = pi_log::init();

    // 初始化trace
    // let _ = pi_log::init_trace();

    //匹配启动时的选项和参数
    let matches = App::new("Pi Serv Main")
        .version("0.2.13")
        .author("YiNeng <yineng@foxmail.com>")
        .arg(
            Arg::with_name("INIT_HEAP_SIZE") //虚拟机初始堆大小
                .short("I")
                .long("INIT_HEAP_SIZE")
                .value_name("Mbytes")
                .help("Sets init vm heap size")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("MAX_HEAP_SIZE") //虚拟机最大堆大小
                .short("H")
                .long("MAX_HEAP_SIZE")
                .value_name("Mbytes")
                .help("Sets max vm heap size")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("MAX_SEMI_SIZE") //虚拟机最大半空间大小
                .short("S")
                .long("MAX_SEMI_SIZE")
                .value_name("Mbytes")
                .help("Sets max vm semi space size")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("MAX_OLD_SIZE") //虚拟机最大旧空间大小
                .short("O")
                .long("MAX_OLD_SIZE")
                .value_name("Mbytes")
                .help("Sets max vm old space size")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("WORK_VM_MULTIPLE") //工作虚拟机倍数
                .short("W")
                .long("WORK_VM_MULTIPLE")
                .value_name("Multiple")
                .help("Sets multiple of work vm amount")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("VM_CPU_COUNT") //cpu数量
                .long("VM_CPU_COUNT")
                .value_name("Multiple")
                .help("Sets multiple of cpu amount")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("EXT_VM_COUNT") //扩展虚拟机数量
                .short("E")
                .long("EXT_VM_COUNT")
                .value_name("Multiple")
                .help("Sets multiple of ext vm amount")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("BASE") //基础版本
                .long("BASE")
                .help("Set the current startup to the base version(Reduce the number of workvm and extvm)")
        )
        .arg(
            Arg::with_name("DEBUG") //工作虚拟机调试模式
                .short("D")
                .long("DEBUG")
                .value_name("Port")
                .help("Enable debug work vm on port")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("DEBUG_BREAK") //工作虚拟机允许调试暂停
                .short("B")
                .long("DEBUG_BREAK")
                .help("Enable debug break work vm on port"),
        )
        .arg(
            Arg::with_name("CONSOLE") //工作虚拟机控制台模式
                .short("C")
                .long("CONSOLE")
                .help("Enable with console for work vm"),
        )
        .arg(
            Arg::with_name("PROFILING") //虚拟机剖析模式
                .short("P")
                .long("PROFILING")
                .help("Enable profiling vm"),
        )
        .arg(
            Arg::with_name("LITE") //打开虚拟机Lite模式
                .short("L")
                .long("LITE")
                .help("Enable vm lite mode"),
        )
        .arg(
            Arg::with_name("JITLESS") //禁止虚拟机jit
                .short("J")
                .long("JITLESS")
                .help("Disable vm jit"),
        )
        .arg(
            Arg::with_name("trace-gc") //打开gc跟踪
                .long("trace-gc")
                .help("Output in stdout following each garbage collection"),
        )
        .arg(
            Arg::with_name("trace-gc-detail") //打开gc详细跟踪
                .long("trace-gc-detail")
                .help("Output detail info in stdout following each garbage collection"),
        )
        .arg(
            Arg::with_name("trace-alloc") //打开分配跟踪
                .long("trace-alloc")
                .help("Output in stdout following each allocate"),
        )
        .arg(
            Arg::with_name("promise-statistics") //打开Promise统计
                .long("promise-statistics")
                .help("Enable promise statistics"),
        )
        .arg(
            Arg::with_name("init-file") // pi_pt入口文件
                .short("i")
                .long("init-file")
                .value_name("init-file")
                .help("pi_pt entry file")
                .takes_value(true),
        )
        .arg(
            Arg::with_name("projects") // 要启动的项目
                .short("p")
                .long("projects")
                .value_name("projects")
                .help("projectw to launch")
                .multiple(true)
                .takes_value(true),
        )
        .get_matches();

    //初始化V8环境，并启动初始虚拟机
    let mut init_heap_size = 16 * 1024 * 1024; //默认虚拟机初始堆大小为16MB
    if let Some(size) = matches.value_of("INIT_HEAP_SIZE") {
        match size.parse::<usize>() {
            Err(e) => panic!("Init v8 env failed, reason: {:?}", e),
            Ok(num) => {
                if num.is_power_of_two() {
                    init_heap_size = num * 1024 * 1024;
                }
            }
        }
    }

    let mut max_heap_size = 8096 * 1024 * 1024; //默认虚拟机最大堆大小为8GB
    if let Some(size) = matches.value_of("MAX_HEAP_SIZE") {
        match size.parse::<usize>() {
            Err(e) => panic!("Init v8 env failed, reason: {:?}", e),
            Ok(num) => {
                if num.is_power_of_two() {
                    max_heap_size = num * 1024 * 1024;
                }
            }
        }
    }

    let mut max_semi_size = None; //默认虚拟机最大半空间大小
    if let Some(size) = matches.value_of("MAX_SEMI_SIZE") {
        match size.parse::<usize>() {
            Err(e) => panic!("Init v8 env failed, reason: {:?}", e),
            Ok(num) => {
                max_semi_size = Some(num);
            }
        }
    }

    let mut max_old_size = None; //默认虚拟机最大旧空间大小
    if let Some(size) = matches.value_of("MAX_OLD_SIZE") {
        match size.parse::<usize>() {
            Err(e) => panic!("Init v8 env failed, reason: {:?}", e),
            Ok(num) => {
                max_old_size = Some(num);
            }
        }
    }

    let mut debug_port: u16 = 0;
    if let Some(value) = matches.value_of("DEBUG") {
        match value.parse::<u16>() {
            Err(e) => {
                panic!("Bind debug listene port failed, reason: {:?}", e);
            }
            Ok(port) => {
                if port > 1024 {
                    debug_port = port;
                } else {
                    panic!(
                        "Bind debug listene port failed, port: {}, reason: invalid port",
                        port
                    );
                }
            }
        }
    }

    let mut debug_break = false;
    if let Some(_) = matches.index_of("DEBUG_BREAK") {
        //允许虚拟机调试暂停
        debug_break = true;
    }

    let mut is_profiling = false;
    if let Some(_) = matches.index_of("PROFILING") {
        //允许剖析虚拟机
        is_profiling = true;
    }

    let mut is_lite = false;
    if let Some(_) = matches.index_of("LITE") {
        //打开虚拟机lite模式
        is_lite = true;
    }

    let mut is_jitless = false;
    if let Some(_) = matches.index_of("JITLESS") {
        //允许剖析虚拟机
        is_jitless = true;
    }

    let mut is_trace_gc = false;
    if let Some(_) = matches.index_of("trace-gc") {
        //允许gc跟踪
        is_trace_gc = true;
    }

    let mut is_trace_gc_detail = false;
    if let Some(_) = matches.index_of("trace-gc-detail") {
        //允许gc跟踪
        is_trace_gc_detail = true;
    }

    let mut is_trace_alloc = false;
    if let Some(_) = matches.index_of("trace-alloc") {
        //允许分配跟踪
        is_trace_alloc = true;
    }
    // if let Some(_) = matches.index_of("promise-statistics") {
    //     //允许Promise统计
    //     vm_builtin::promise::enable_global_promise_hooks();
    // }

    //允许Promise统计
    vm_builtin::promise::enable_global_promise_hooks();

    // 创建网络运行时
    // let len = get_physical();
    // let mut new_work_rts = Vec::with_capacity(len);
    // for _ in 0..len {
    //     let rt: LocalTaskRuntime<()> = AsyncRuntimeBuilder::default_local_thread(None, None);
    //     new_work_rts.push(rt);
    // }
    // let new_work_rts = Arc::new(new_work_rts);

    let (init_heap_size, max_heap_size, debug_port) = init_v8_env::<VmTaskPool, VmTaskPool>(
        0,
        false,
        init_heap_size,
        max_heap_size,
        max_semi_size,
        max_old_size,
        debug_port,
        is_profiling,
        is_lite,
        is_jitless,
        is_trace_gc,
        is_trace_gc_detail,
        is_trace_alloc,
    );

    let matches_copy = matches.clone();
    if let Err(e) = MAIN_ASYNC_RUNTIME.spawn(async move {
        async_main(
            matches_copy,
            init_heap_size,
            max_heap_size,
            debug_port,
            debug_break,
        )
        .await;
    }) {
        panic!("Spawn async main task failed, reason: {:?}", e);
    }

    let (send, recv) = channel();

    // 动态加载项目的时候，使用主线程运行
    *LOAD_PROJECT_SENDER.clone().lock() = Some(send);

    //主线程循环
    while MAIN_RUN_STATUS.load(Ordering::Relaxed) {
        if let Ok((
            init_heap_size,
            max_heap_size,
            debug_port,
            debug_break,
            work_vm_count,
            init_exec_path,
            is_console,
            project,
            ext_vm_count,
            rsp_send,
        )) = recv.try_recv()
        {
            if let Err(e) = MAIN_ASYNC_RUNTIME.spawn(async move {
                init_main(
                    init_heap_size,
                    max_heap_size,
                    debug_port,
                    debug_break,
                    work_vm_count,
                    init_exec_path,
                    is_console,
                    project,
                    ext_vm_count,
                    rsp_send,
                )
                .await;
            }) {
                panic!("Spawn async main task failed, reason: {:?}", e);
            }
        }
        //推动主线程异步运行时
        let start_time = Instant::now();
        if let Err(e) = MAIN_ASYNC_RUNNER.run() {
            panic!("Main loop failed, reason: {:?}", e);
        }
        let run_time = Instant::now() - start_time;

        if let Some(remaining_interval) =
            Duration::from_millis(*MAIN_UNCONDITION_SLEEP_TIMEOUT).checked_sub(run_time)
        {
            //本次运行少于循环间隔，则休眠剩余的循环间隔，并继续执行任务
            thread::sleep(remaining_interval);
        }
    }
}

/*
* 异步执行入口，退出时不会中止主线程
*/
async fn async_main(
    matches: ArgMatches<'static>,
    init_heap_size: Option<usize>,
    max_heap_size: Option<usize>,
    debug_port: Option<u16>,
    debug_break: bool,
) {
    let mut db_path = "./db".to_string();
    if let Ok(var) = env::var("DB_PATH") {
        db_path = var.parse().unwrap_or("./db".to_string());
    }
    // 加载native funtion
    register_ext_functions();
    // 初始化数据库
    init_db(DB_ASYNC_RUNTIME.clone(), db_path).await;
    // set_store_runtime(FILES_ASYNC_RUNTIME.clone()).await;
    // 注册文件异步运行时
    set_file_async_runtime(FILES_ASYNC_RUNTIME.clone());
    set_pi_serv_lib_file_runtime(FILES_ASYNC_RUNTIME.clone());
    set_pi_serv_lib_db_runtime(DB_ASYNC_RUNTIME.clone());
    set_pi_serv_lib_main_async_runtime(MAIN_ASYNC_RUNTIME.clone());
    // 注册pi_serv方法
    reg_pi_serv_handle();
    // 注册pi_serv_builtin运行时
    set_external_async_runtime(MAIN_ASYNC_RUNTIME.clone());
    // 设置env
    set_current_env();
    // 设置logfilr大小
    // set_logfile_size();

    let is_console = if let None = matches.index_of("CONSOLE") {
        //不打开控制台，则忽略
        false
    } else {
        true
    };

    let init_exec_path = matches.value_of("init-file").unwrap().to_string();
    let projs = match matches.values_of("projects") {
        Some(p) => p
            .map(|s| s.to_string().replace("\\", "/"))
            .collect::<Vec<String>>(),
        None => vec![],
    };

    let mut cpus = get_physical();
    if let Some(value) = matches.value_of("VM_CPU_COUNT") {
        match value.parse::<usize>() {
            Err(e) => {
                panic!("VM_CPU_COUNT failed, reason: {:?}", e);
            }
            Ok(count) => {
                cpus = count;
            }
        }
    }
    info!("==========cpus:{:?}", cpus);
    let mut work_vm_count: usize = 2 * cpus + 2 + 2; //默认工作虚拟机数量为本地cpu物理核数的2倍, 2个：mqttlistener、httplistener、2个定时服务
    if let Some(value) = matches.value_of("WORK_VM_MULTIPLE") {
        match value.parse::<usize>() {
            Err(e) => {
                panic!("Init work vm failed, reason: {:?}", e);
            }
            Ok(0) => {
                work_vm_count = 1 + 2 + 2; // 用户虚拟机、mqttlistener、httplistener、定时服务
            }
            Ok(count) => {
                work_vm_count = cpus * count + 2 + 2;
            }
        }
    }

    env::set_var("WORK_VM_COUNT", work_vm_count.to_string());

    let mut ext_vm_count: usize = 2 * cpus; //默认工作虚拟机数量
    if let Some(value) = matches.value_of("EXT_VM_COUNT") {
        match value.parse::<usize>() {
            Err(e) => {
                panic!("Init ext vm failed, reason: {:?}", e);
            }
            Ok(0) => {
                ext_vm_count = 1;
            }
            Ok(count) => {
                ext_vm_count = cpus * count;
            }
        }
    }

    env::set_var("EXT_VM_COUNT", ext_vm_count.to_string());
    // 如果当前启动的是基础版本，减少虚拟机数量
    if let Some(_) = matches.index_of("BASE") {
        work_vm_count = 2 + 2 + 2;
        ext_vm_count = 2;
    }

    let current_dir = env::current_dir().unwrap();
    let current_dir_parent = current_dir.parent().unwrap().to_str().unwrap();

    let path = Path::new(&init_exec_path)
        .iter()
        .filter_map(|x| if x == "." || x == ".." { None } else { Some(x) })
        .map(|x| x.to_str().unwrap())
        .collect::<Vec<&str>>();

    let root: PathBuf = [vec![current_dir_parent], path].concat().iter().collect();
    let project_root = root
        .parent()
        .unwrap()
        .parent()
        .unwrap()
        .to_str()
        .unwrap()
        .replace("\\", "/");

    env::set_var("PROJECTS", &projs.as_slice().join(" "));

    let cur_dir = env::current_dir();

    // 如果没有出现 -p 参数
    if matches.occurrences_of("projects") == 0 {
        env::set_var("PROJECT_ROOT", cur_dir.unwrap().to_str().unwrap());
    } else {
        env::set_var("PROJECT_ROOT", &project_root);
    }
    let mut size = projs.len();
    for project in projs {
        size = size - 1;
        // 在最后一个project中启动控制台
        let is_console = if size == 0 { is_console } else { false };
        let init_exec_path = project.clone() + "/pi_pt/init.js";
        init_main(
            init_heap_size,
            max_heap_size,
            debug_port,
            debug_break,
            work_vm_count,
            init_exec_path,
            is_console,
            project,
            ext_vm_count,
            None,
        )
        .await;
    }

    // 加载监控
    let _ = js_monitor::start();

    // if let Some(start_at) = get_db_collect_time() {
    //     collect_db(FILES_ASYNC_RUNTIME.clone(), start_at);
    // } else {
    //     warn!("no database collect time specified");
    // }
}

// 执行平台初始化
pub async fn init_pt(worker: vm::Vm<VmTaskPool>, rt: MultiTaskRuntime) -> Result<()> {
    let cid = worker.alloc_context_id();
    let context = worker.new_context(None, cid, None).await.unwrap();
    let _ = worker
        .execute(
            context,
            "pi_pt/init/init.js",
            r#"self.Module.require("pi_pt/init/init.js");"#,
        )
        .await;
    while !(*PI_PT_INIT_STATE.read().await).0 {
        rt.timeout(1000).await;
        // thread::sleep(Duration::from_millis(1000));
    }
    if let Some(error) = &(*PI_PT_INIT_STATE.read().await).1 {
        Err(Error::new(ErrorKind::Other, error.clone().to_string()))
    } else {
        Ok(())
    }
}

// // 执行mfa
// pub async fn init_mfa(worker: vm::Vm) {
//     let file = std::fs::read_to_string("../dst_server/ptconfig.json").unwrap();
//     let json = parse(&file).unwrap();
//     let mfas = &json["MFA"];
//     let mut execs = vec![];
//     for m in mfas.members() {
//         debug!(
//             "mfa exec module = {:?}, func = {:?}, args = {:?}",
//             m["module"], m["function"], m["args"]
//         );
//         execs.push(format!(
//             r#"
//             var JsDb  = self.Module.require('pi_pt/native/pi_serv_lib/js_db.js');
//             var guid = JsDb.GuidGen.new(1, 1);
//             var mgr = JsDb.Mgr.new(guid.get_self());
//             self.mgr = mgr;
//             var envMod = Module.require('pi_utils/lang/env');
//             self.env = new envMod.Env();
//             env.other = new Map();
//             var mgrMod = Module.require('pi_pt/db/mgr');
//             if (!env.dbMgr) env.dbMgr = new mgrMod.Mgr(mgr);
//             Module.require("{}");
//             var f = Module.modules["{}"].exports.{}.apply(Module.modules["{}"].exports.{}, {});
//             if (f instanceof Promise) f.then(() => console.log("mfa init finished")).catch(e => console.log("mfa init failed", e.toString()));"#,
//             m["module"], m["module"], m["function"], m["module"], m["function"], m["args"]
//         ));
//     }

//     let cid = worker.alloc_context_id();
//     let context = worker.new_context(None, cid, None).await.unwrap();

//     for exec in execs {
//         let _ = worker.execute(context, "", &exec).await;
//     }
// }

// 创建灰度版本
pub(crate) fn get_gray_last_v() -> usize {
    let mut gray_ver: usize = 100;
    let grays = GRAY_MGR.read().all_grays();
    if let Some(last) = grays.last() {
        gray_ver = ((last / 100) + 1) * 100;
    }
    gray_ver
}
//初始化默认灰度
pub(crate) fn init_default_gray(
    workers: Vec<vm::Vm<VmTaskPool>>,
    ext_vms: Vec<vm::Vm<VmTaskPool>>,
    vm_args: GrayVmArgs,
    gray_ver: usize,
    project: String,
) {
    if let Err(e) = GRAY_MGR.write().add_new_gray(
        gray_ver,
        workers,
        ext_vms,
        vm_args,
        global_db_mgr(),
        project,
    ) {
        panic!("Create default gray failed, reason: {:?}", e);
    }
}

//初始化端口的监听Pid
pub(crate) fn init_listener_pid(gray: usize) {
    for (port, broker_name) in MQTT_PORTS.lock().iter() {
        create_listener_pid(port.clone(), broker_name, gray);
    }
}

//初始化HTTP端口的监听Pid
pub(crate) fn init_http_listener_pid(gray: usize) {
    for (port, host, config) in HTTP_PORTS.lock().iter() {
        create_http_pid(host, port.clone(), gray, config);
    }
}

// 注册虚拟机关心处理的事件
pub(crate) fn reigster_vms_events(workers: &[vm::Vm<VmTaskPool>], is_debug_mode: bool) {
    // 设置虚拟机的事件回调
    for worker in workers {
        let vm = worker.clone();
        let event_handler =
            VmEventHandler::new(MAIN_ASYNC_RUNTIME.clone(), move |event, vid| match event {
                VmEventValue::CreatedContext(context) => {
                    debug!(
                        "Vm event handler: VmEventValue::CreatedContext, vid = {:?}, cid = {:?}",
                        vid, context.0
                    );

                    // 非调试模式下才需要重新require热更过的文件
                    // if !is_debug_mode {
                    //     // 获取已经热更新过的文件
                    //     let mut hotfixed_files = HOTFIX_FILES
                    //         .lock()
                    //         .iter()
                    //         .map(|(key, val)| (key.clone(), val.clone()))
                    //         .collect::<Vec<(String, usize)>>();
                    //     // 按版本号从小到大排序
                    //     hotfixed_files.sort_by(|a, b| a.1.cmp(&b.1));

                    //     for (path, _) in hotfixed_files {
                    //         let vm = vm.clone();
                    //         let _ =
                    //             MAIN_ASYNC_RUNTIME.spawn(MAIN_ASYNC_RUNTIME.alloc(), async move {
                    //                 if let Err(e) = hotfix_load(vid, vm, context, path).await {
                    //                     warn!("hotfix error: {:?}", e);
                    //                 }
                    //             });
                    //     }
                    // }

                    VID_CONTEXTS
                        .entry(vid)
                        .and_modify(|v| {
                            v.insert(context);
                        })
                        .or_insert(HashSet::from([context]));

                    // if is_debug_mode {
                    //     let vm = GRAY_MGR.read().vm_instance(0, vid).unwrap();
                    //     let vm_copy = vm.clone();
                    //     vm.spawn_task(async move {
                    //         let source =
                    //             read_init_source("../dst_server/pi_pt/init.js".to_string()).await;
                    //         if let Err(e) =
                    //             vm_copy.execute(context, "init.js", source.as_ref()).await
                    //         {
                    //             panic!(e);
                    //         }
                    //     });
                    // }
                }

                VmEventValue::RemovedContext(context) => {
                    debug!(
                        "Vm event handler: VmEventValue::RemovedContext, vid = {:?}, cid = {:?}",
                        vid, context.0
                    );
                    pid_close_count(vid, context);
                    VID_CONTEXTS.entry(vid).and_modify(|v| {
                        v.retain(|ctx| *ctx != context);
                    });
                }
            });
        worker.register_event_handler(VmEvent::CreatedContext, event_handler.clone());
        worker.register_event_handler(VmEvent::RemovedContext, event_handler);
    }
}

// 热更加载
async fn hotfix_load(
    vid: usize,
    vm: vm::Vm<VmTaskPool>,
    context: ContextHandle,
    path: String,
) -> Result<()> {
    if let Ok(Some(func)) = vm
        .get_property(context.clone(), "self.Module.require")
        .await
    {
        if let Some(p) = vm
            .to_js_value(context.clone(), stringify(path.clone()))
            .await?
        {
            if let Some(dir) = vm.to_js_value(context.clone(), stringify("")).await? {
                if let Some(force) = vm.to_js_value(context.clone(), stringify(true)).await? {
                    vm.call(context.clone(), &func, vec![p.clone(), dir, force])
                        .await?;
                }
            }
        }
        Ok(())
    } else {
        Err(Error::new(ErrorKind::Other, "get Module.require error"))
    }
}

// 通过环境变量控制是否启动热更新
fn enable_hotfix(gray: usize, project: String) {
    // let p = if let Ok(p) = env::var("PROJECTS") {
    //     p
    // } else {
    //     String::from("../dst_server")
    // };
    if env::var("ENABLE_HOTFIX").is_ok() {
        info!("Start listen hotfix...");
        hotfix_listen_backend(gray, project);
        hotfix_listen_frontend();
    }
}

//初始化控制台
async fn init_console(
    workers: &Vec<(Arc<AtomicBool>, vm::Vm<VmTaskPool>)>,
) -> Option<(vm::Vm<VmTaskPool>, ContextHandle)> {
    let (_, worker) = &workers[0];
    let context = worker
        .new_context(None, worker.alloc_context_id(), None)
        .await
        .unwrap();

    //初始化控制台运行时
    let pool = VmTaskPool::default();
    let console_runner = SingleTaskRunner::new(pool);
    let console_runtime = console_runner.startup().unwrap();
    thread::Builder::new()
        .name("Pi-Exec-Console".to_string())
        .stack_size(2 * 1024 * 1024)
        .spawn(move || {
            loop {
                //推动控制台异步运行时
                let start_time = Instant::now();
                if let Err(e) = console_runner.run() {
                    panic!("Main loop failed, reason: {:?}", e);
                }
                let run_time = Instant::now() - start_time;

                if let Some(remaining_interval) = Duration::from_millis(10).checked_sub(run_time) {
                    //本次运行少于循环间隔，则休眠剩余的循环间隔，并继续执行任务
                    thread::sleep(remaining_interval);
                }
            }
        })
        .unwrap();

    //构建控制台，并注册
    let console_shell: ConsoleShell<SingleTaskRuntime<(), VmTaskPool>, VmTaskPool, VmTaskPool> =
        ConsoleShellBuilder::<SingleTaskRuntime<(), VmTaskPool>, VmTaskPool, VmTaskPool>::new(
            console_runtime,
            worker.clone(),
            context,
        )
        .buffer_stdout()
        .enable_color_stdout()
        .enable_color_stderr()
        .set_title("PiServ Console Shell")
        .build();
    *CONSOLE_SHELL.write() = Some(console_shell);

    Some((worker.clone(), context))
}

// 环境变量设置
fn set_current_env() {
    if env::var("CURRENT_LIMIT").is_ok() {
        env::set_var("current", "true");
    } else {
        env::set_var("current", "false");
    }
}

/// pid关闭计数
/// vid 虚拟机ID
/// context 虚拟机上下文
fn pid_close_count(vid: usize, _context: ContextHandle) {
    let mut map = AUTO_GC.write();
    if let Some(atom) = map.get(&vid) {
        let old = atom.fetch_add(1, Ordering::SeqCst);
        let mut count = 10;
        if let Ok(var) = env::var("AUTO_GC_COUNT") {
            count = var.parse().unwrap_or(10);
        }
        // 是否需要GC
        if old >= count {
            // 重置计数
            atom.store(1, Ordering::Relaxed);

            let pid = Pid(vid, 0);
            let mut msgs = Vec::with_capacity(1);
            msgs.push(ProcessMsg::String("gc".to_string()));
            // listenerPID发送消息
            send_to_process::<VmTaskPool, VmTaskPool>(None, pid, ProcessMsg::Array(msgs));
        }
    } else {
        map.insert(vid, AtomicUsize::new(1));
    }
}

// 获取数据库表整理时间 0~23
// fn get_db_collect_time() -> Option<u32> {
//     let file = std::fs::read_to_string("../dst_server/ptconfig.json").unwrap();
//     let json = parse(&file).unwrap();
//     let time = &json["collect_time"];
//     time.as_u32()
// }

// 设置logfile文件大小
// fn set_logfile_size() -> () {
//     let mut size = 200;
//     if let Ok(var) = env::var("LOGFILE_SIZE") {
//         size = var.parse().unwrap_or(200);
//     }
//     if (size > 0 && size < 1024) {
//         LOG_FILE_SIZE.store(size, Ordering::SeqCst);
//     }
// }
